extends KinematicBody

onready var ownId:int = int(self.name)

######## MOVEMENT AND POSITION VARS ########

onready var collisionShape:CollisionShape = $collisionShape
var bodyRotation:int = 1

# Speed values
const MAX_SPEED:float = 8.0
var motion:float = 0
var isSprinting:bool = false

# Acceleration values
const ACCEL:float = 8.5
const MVMNT_MOMENTUM:float = 10.0
const THROW_MOMENTUM:float = 2.0

# Jump and fall values
const JUMP_SPEED:float = 16.0
var isJumping:bool = false

var maxJump:int = 2         # Maximum number of jump that a character can do
var currentJump:int = 0     # Actual number of jump
var lastJumpId:int = 0      # Detect a new jump

# TODO: unreal, we may want to stick that to the physical value
const GRAVITY:float = -28.0

# Velocity updated continuously by move_and_slide
var vel:Vector3 = Vector3(0,0,0)

# move_and_slide need that
const FLOOR_NORMAL:Vector3 = Vector3(0,1,0)

######### ATTACKS VARS ########

onready var lastStrikerId:int = ownId;

var isAttacking:bool = false
onready var attackTimer:Tween = $attackTween

# attack type
enum {
	NONE = 0,
	PUNCH = 1,
	KICK = 2
}
enum {
	ID_PUNCH_1 = 0,
	ID_PUNCH_2 = 1,
	ID_KICK_1 = 2,
	ID_KICK_2 = 3,
	ID_KICK_JUMP_SIDE = 4
}
var lastPunchId:int = 0     # Detect a new punch (of any type)
var idleTime:float = 0      # Stun when hit + prevent spam

# Mouse
var attackDirection:int
enum {
	MOUSE_LEFT = 0,
	MOUSE_UP = 1,
	MOUSE_RIGHT = 2,
	MOUSE_DOWN = 3
}

# Punch attack parameters
onready var punchHitArea:Area = $punchHitArea
var punchAttackDmg:float = 5.0
var punchAttackDist:float = 1.2
var punchAttackWait:float = 0.6
var punch1AttackDuration:float = 0.1
var punch2AttackDuration:float = 0.4
# General direction (normalized anyway)
var punchAttackDirection:Vector3 = Vector3(1,1.5,0)
# Strength of the attack (multiply the normalized vector)
var punchAttackStrength:float = 0.5

# Kick attack parameters
onready var kickHitArea:Area = $kickHitArea
var kickAttackDmg:float = 10.0
var kick1AttackDist:float = 1.7
var kick2AttackDist:float = 1.4
var kickJumpSideAttackDist:float = 1.2
var kickAttackWait:float = 0.75
var kick1AttackDuration:float = 0.2
var kick2AttackDuration:float = 0.3
var kickJumpSideAttackDuration:float = 0.3
var kickAttackDirection:Vector3 = Vector3(1.2,1,0)
var kickAttackStrength:float = 1

# Health points related values
var hp:int = 0

######## ANIMATIONS ########

var animSequence:int = 0

enum {
	ANIM_CONTROLLED_FALL = -1,
	ANIM_IDLE = 0,
	ANIM_RUN = 1,
	ANIM_JUMP = 2,
	ANIM_PUNCH_1 = 3,
	ANIM_PUNCH_2 = 4,
	ANIM_KICK_1 = 5,
	ANIM_KICK_2 = 6,
	ANIM_JUMP_KICK_SIDE = 7,
	ANIM_JUMP_KICK_UP = 8,
	ANIM_JUMP_KICK_DOWN = 9,
	ANIM_TURN = 10,
	ANIM_IMPACT_1 = 11,
	ANIM_IMPACT_2 = 12,
	ANIM_IMPACT_JUMP = 13,
	ANIM_BIG_IMPACT = 14
}

######## FUNCTIONS ########

# called by the engine
func _physics_process(delta:float):
	processMovement(delta)
	broadcastMovement()


# Called from the game script to update the vars
func getPlayerInputs(movementInput:int, jumpId:int, attackTypeInput:int, mouseX:int, mouseY:int):
	# Prevent the player from moving when attacking
	if isAttacking:
		return

	motion = movementInput

	# Differentiate a new jump from a reemitted one
	if lastJumpId != jumpId:
		lastJumpId = jumpId
		isJumping = true
	
	# Execute attack
	if attackTypeInput != NONE:
		attackDirection = MOUSE_UP
		if (!self.is_on_floor() && mouseY < 0 && mouseX < mouseY && mouseX > -mouseY):
			attackDirection = MOUSE_DOWN
		elif (mouseX < 0 && mouseX <= mouseY):
			attackDirection = MOUSE_LEFT
			bodyRotation = -1
			set_rotation_degrees(Vector3(0, 180, 0))
			broadcastTurn(bodyRotation)
		elif (mouseX > 0 && mouseX >= -mouseY):
			attackDirection = MOUSE_RIGHT
			bodyRotation = 1
			set_rotation_degrees(Vector3(0, 0, 0))
			broadcastTurn(bodyRotation)
		processAttack(attackTypeInput)


func processMovement(delta:float):
	# Set the side faced by the character
	if motion > 0:
		if !isAttacking:
			bodyRotation = 1
			set_rotation_degrees(Vector3(0, 0, 0))
			broadcastTurn(bodyRotation)
		if self.is_on_floor():
			broadcastAnimation(ANIM_RUN)
		else:
			broadcastAnimation(ANIM_CONTROLLED_FALL)
	elif motion < 0:
		if !isAttacking:
			bodyRotation = -1
			set_rotation_degrees(Vector3(0, 180, 0))
			broadcastTurn(bodyRotation)
		if self.is_on_floor():
			broadcastAnimation(ANIM_RUN)
		else:
			broadcastAnimation(ANIM_CONTROLLED_FALL)
	else:
		if self.is_on_floor():
			broadcastAnimation(ANIM_IDLE)
		else:
			broadcastAnimation(ANIM_CONTROLLED_FALL)


	if self.is_on_floor():
		# Reset the number of jumps left
		currentJump = 0
		# Reset the last striker
		lastStrikerId = ownId
	elif currentJump == 0 && !isJumping:
		# Initial jump prevented
		currentJump = 1

	# Jump attempts computation
	if isJumping:
		isJumping = false
		if currentJump < maxJump:
			broadcastAnimation(ANIM_JUMP)
			currentJump += 1
			# Consider the gravity when jumping
			vel.y = JUMP_SPEED - delta*GRAVITY

	# Consider the gravity
	vel.y += delta*GRAVITY

	# Speed of the player
	var target:Vector3 = Vector3(motion, 0, 0)

	target *= MAX_SPEED

	# Acceleration of the player if he is moving horizontally
	var accel:float
	var hvel:Vector3 = Vector3(vel.x, 0, 0)
	if target.dot(hvel)>0:
		accel = ACCEL
	elif !is_on_floor()&&!isJumping:
		accel = THROW_MOMENTUM
	else:
		accel = MVMNT_MOMENTUM

	# Interpolating speed and acceleration through time and getting it
	vel.x = hvel.linear_interpolate(target, accel * delta).x

	# Update the velocity after the collisions (physical engine ftw)
	vel = self.move_and_slide(vel, FLOOR_NORMAL)

#	if get_slide_count() > 0:
#      var collision = get_slide_collision(0)
#      if collision != null:
#         vel = vel.bounce(collision.normal)


# Send the position of the player
func broadcastMovement():
	var position = self.get_translation()
	position.z = 0
	rpc("getRemoteMovement", position)


# Check and trigger the type of attack requested
func processAttack(attackType:int):
	# Turn off the other attacks input for the time being
	isAttacking = true

	# Use the right function
	match attackType:
		PUNCH:
			if self.is_on_floor():
				if (!animSequence):
					broadcastAnimation(ANIM_PUNCH_1)
					animSequence = 1
					attack(ID_PUNCH_1)
				else:
					broadcastAnimation(ANIM_PUNCH_2)
					animSequence = 0
					attack(ID_PUNCH_2)
			else:
				broadcastAnimation(ANIM_JUMP_KICK_SIDE)
				attack(ID_KICK_JUMP_SIDE)
		KICK:
			if self.is_on_floor():
				if (!animSequence):
					broadcastAnimation(ANIM_KICK_1)
					animSequence = 1
					attack(ID_KICK_1)
				else:
					broadcastAnimation(ANIM_KICK_2)
					animSequence = 0
					attack(ID_KICK_2)
			else:
				broadcastAnimation(ANIM_JUMP_KICK_SIDE)
				attack(ID_KICK_JUMP_SIDE)


func attack(idAttack:int):
	var hitArea:Area
	var distance:float
	var wait:float
	var duration:float

	if (idAttack == ID_PUNCH_1 || idAttack == ID_PUNCH_2):
		hitArea = punchHitArea
		distance = punchAttackDist
		wait = punchAttackWait
		if (idAttack == ID_PUNCH_1):
			duration = punch1AttackDuration
		else:
			duration = punch2AttackDuration
	else:
		hitArea = kickHitArea
		wait = kickAttackWait
		if (idAttack == ID_KICK_1):
			distance = kick1AttackDist
			duration = kick1AttackDuration
		elif (idAttack == ID_KICK_2):
			distance = kick2AttackDist
			duration = kick2AttackDuration
		else:
			distance = kickJumpSideAttackDist
			duration = kickJumpSideAttackDuration
	
	# Set active the hitbox of the attack, its visual feedback on server
	hitArea.set_monitoring(true)
	hitArea.set_visible(true)

	# Prepare the translation as an interpolation over time (end position, duration and function used)
	attackTimer.interpolate_property(hitArea, "translation:x", 0, distance, duration, Tween.TRANS_LINEAR, Tween.EASE_OUT)
	attackTimer.start()
	yield(attackTimer, "tween_completed")
	hitArea.set_monitoring(false)
	attackTimer.interpolate_property(hitArea, "translation:x", distance, 0, wait-duration, Tween.TRANS_LINEAR, Tween.EASE_OUT)
	attackTimer.start()

	# Once the movement is finished, turn off the hitbox and visual feedback
	yield(attackTimer, "tween_completed")
	isAttacking = false
	hitArea.set_visible(false)


# Called when detected by an attack's hitbox
func hurt(damages:int, sourceId:int, direction:Vector3, strength:float):
	# Play hurt animation
	if self.is_on_floor():
		if (!animSequence):
			broadcastAnimation(ANIM_IMPACT_1)
		else:
			broadcastAnimation(ANIM_IMPACT_2)
	else:
		broadcastAnimation(ANIM_IMPACT_JUMP)

	# Add damages to your bar
	hp += damages

	# Projection: normalized vector * direction * strength added to the velocity of the player
	vel += strength*hp*direction

	# Update the last played who hit you
	lastStrikerId = sourceId

	# Feedback on the player's client
	broadcastHp()


# Send Hp
func broadcastHp():
	rpc("hurt", hp)

# Make player die
func die():
	rpc("die", lastStrikerId)
	self.queue_free()

# Make character turn
func broadcastTurn(direction:int):
	rpc("turn",direction)
	
# Send the type action of the player
func broadcastAnimation(animationType:int):
	rpc("getAnimationTypePlay",animationType)

######## SIGNALS ########

# General function on attack's hitboxes
func _on_generalHitArea_entered(damageVar:int, directionVar:Vector3,
strengthVar:float, body:Node):
	# Don't hurt yourself
	if int(body.name)==ownId:
		return

	# Normalized projection vector
	var direction:Vector3 = directionVar.normalized()

	# Rotate the vector depending on the player orientation
	direction.x*=bodyRotation

	print("body named " + body.name + " hit: punch")

	# Call the function on the affected player
	body.hurt(damageVar, ownId, direction, strengthVar)


# punch hit has landed on something
func _on_punchHitArea_body_entered(body:Node):
	_on_generalHitArea_entered(punchAttackDmg, punchAttackDirection,
	punchAttackStrength, body)


# kick hit has landed on something
func _on_kickHitArea_body_entered(body:Node):
	_on_generalHitArea_entered(kickAttackDmg, kickAttackDirection,
	kickAttackStrength, body)
